/*
 *  RexxExpat - Version 1.0 - 15th January, 2001
 *  Copyright (C) 2001 Dominik Stein <dominik_stein@gmx.de>
 *
 *  This library wraps the C-XML-parser 'Expat' so it can be used 
 *  from a Rexx environment. 
 *
 */

#define INCL_RXFUNC					/* external functions support */
#define INCL_RXSHV					/* shared variable support    */
#define INCL_RXSYSEXIT				/* system exit support        */

#include <stdio.h>					/* using sprintf         */
#include "xmlparse.h"				/* include expat	 API */
#include "rexxsaa.h"				/* include RexxSAA	 API */
#include "RexxExpat.h"				/* include RexxExpat API */

/* ********************************************************************* *
	Load/Unload RexxExpat Functions
 * ********************************************************************* *
	exported functions:
		APIRET APIENTRY ExpatLoadFuncs (APIINTERFACE);
		APIRET APIENTRY ExpatDropFuncs (APIINTERFACE);
 * ********************************************************************* */

/* functions to be exported */
static PSZ RexxExpatFuncTable[] =
{
	/* Load/Unload RexxExpat */
	/* "ExpatLoadFuncs", */
	"ExpatDropFuncs",
	
	/* Variable Pool Function */
	"ExpatSetVariablePool",
	"ExpatGetVariablePool",
	"ExpatFreeVariablePool",
	
	/* Parser Creation */
	"ExpatParserCreate",
	"ExpatParserCreateNS",
	"ExpatExternalEntityParserCreate",
	"ExpatParserFree",
	
	/* Parsing */
	"ExpatParse",
	
	/* Handler Setting */
	"ExpatSetElementHandler",
	"ExpatSetCharacterDataHandler",
	"ExpatSetProcessingInstructionHandler",
	"ExpatSetCommentHandler",
	"ExpatSetDoctypeDeclHandler",
	"ExpatSetCdataSectionHandler",
	"ExpatSetDefaultHandler",
	"ExpatSetDefaultHandlerExpand",
	"ExpatSetInternalParsedEntityDeclHandler",
	"ExpatSetExternalParsedEntityDeclHandler",
	"ExpatSetExternalEntityRefHandler",
	"ExpatSetUnknownEncodingHandler",
	"ExpatSetUnknownEncodingConverter",
	"ExpatSetNamespaceDeclHandler",
	"ExpatSetUnparsedEntityDeclHandler",
	"ExpatSetNotationDeclHandler",
	"ExpatSetNotStandaloneHandler",
	
	/* Parse Position and Error Reporting Functions */
	"ExpatGetErrorCode",
	"ExpatErrorString",
	"ExpatGetCurrentByteIndex",
	"ExpatGetCurrentLineNumber",
	"ExpatGetCurrentColumnNumber",
	
	/* Miscellaneous Functions */
	"ExpatSetBase",
	"ExpatGetBase",
	"ExpatGetSpecifiedAttributeCount",
	"ExpatGetIdAttributeIndex",
	"ExpatGetCurrentByteCount",
	"ExpatSetEncoding",
	"ExpatSetParamEntityParsing"
};

APIRET APIENTRY ExpatLoadFuncs (APIINTERFACE)
/* loads every functions in RexxExpatFuncTable[] 
	no arguments, no return codes */ 
{
	int rc, i;
	for (i = 0; i < sizeof(RexxExpatFuncTable)/sizeof(PSZ); i++)
		if ((rc = RexxRegisterFunctionDll (RexxExpatFuncTable[i], DLLNAME, RexxExpatFuncTable[i])) != RXFUNC_OK)
			return rc;
	return 0;
}

APIRET APIENTRY ExpatDropFuncs (APIINTERFACE)
/* unloads every functions in RexxExpatFuncTable[] */ 
{
	int i;
	for (i = 0; i < sizeof(RexxExpatFuncTable)/sizeof(PSZ); i++)
		RexxDeregisterFunction (RexxExpatFuncTable[i]);
	return 0;
}

/* ********************************************************************* *
	Rexx Variable Pool Functions
 * ********************************************************************* *
	helper functions:
		APIRET _ExpatFreeRexxVariablePool (PSHVBLOCK *FirstPointer);

	exported functions:
		APIRET APIENTRY ExpatGetVariablePool  (APIINTERFACE);
		APIRET APIENTRY ExpatSetVariablePool  (APIINTERFACE);
		APIRET APIENTRY ExpatFreeVariablePool (APIINTERFACE);
 * ********************************************************************* */

APIRET _ExpatFreeRexxVariablePool (PSHVBLOCK *FirstPointer) {
/* frees list of SHVBLOCKs pointed to by FirstPointer */
/* returns 0 if de-allocation was successful, while on error _REXXEXPATVERIFYDEALLOCATION is processed */
	PSHVBLOCK NextPointer, ListPointer = *FirstPointer;
	while (ListPointer) {
		NextPointer = ListPointer->shvnext;
		if (!(ListPointer->shvret & RXSHV_LVAR)) {
			if (ListPointer->shvname.strptr) {
				_REXXEXPATVERIFYDEALLOCATION (
				REXXFREEMEMORY(ListPointer->shvname.strptr)); }
			if (ListPointer->shvvalue.strptr) {
				_REXXEXPATVERIFYDEALLOCATION (
				REXXFREEMEMORY(ListPointer->shvvalue.strptr)); }
		};
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(ListPointer));
		ListPointer = NextPointer;
	}
	*FirstPointer=NULL;
	return 0;
}

APIRET APIENTRY ExpatFreeVariablePool (APIINTERFACE)
/*	1st argument:	reference to userData
	returns:		0, if successful                 */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* free variable pool - return code must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", _ExpatFreeRexxVariablePool(&userData->shvfirst));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetVariablePool (APIINTERFACE)
/*	1st argument:	reference to userData
	returns:		0, if successful and >127 on fatal error */
{
	/* local variables */
	RexxExpatData	userData;
	PSHVBLOCK		ListPointer;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set variable pool */
	ListPointer = userData->shvfirst;
	while (ListPointer && !(ListPointer->shvret & RXSHV_LVAR)) {
		ListPointer->shvcode = RXSHV_SET;
		ListPointer = ListPointer->shvnext;
	}

	/* execute request - return code must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", RexxVariablePool(userData->shvfirst));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetVariablePool (APIINTERFACE)
/*	1st argument:	reference to userData
	following arguments: names of variable to save 
	if no argument follows, the entire pool is saved
	returns:		0, if successful and >127 on fatal error */
{
	/* local variables */
	RexxExpatData	userData;
	PSHVBLOCK		LastPointer, NewPointer, *SavePointerHere;
	unsigned int 	i;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* init SavePointerHere */
	SavePointerHere = &userData->shvfirst;

	/* any arguments passed? */
	if (Argc>1) {
		/* yes - go to last entry  */
		for (LastPointer = userData->shvfirst; 
			 userData->shvfirst && LastPointer->shvnext; 
			 LastPointer = LastPointer->shvnext, 
				 SavePointerHere = &LastPointer->shvnext); }
	else {
		/* no  - free variable pool */
		_REXXEXPATVERIFYDEALLOCATION (
		_ExpatFreeRexxVariablePool(&userData->shvfirst)); }

	/* append entries */
	for (i=1; (Argc>1 ? i<Argc : (SavePointerHere==&userData->shvfirst || !(NewPointer->shvret & RXSHV_LVAR))); i++) {
		if (Argc>1 && !RXVALIDSTRING(Argv[i])) continue; /* skip invalid arguments */
		_REXXEXPATVERIFYALLOCATION (
		NewPointer = (PSHVBLOCK) REXXALLOCATEMEMORY(sizeof(SHVBLOCK)));
		/* init shvblock */
		NewPointer->shvcode           = (Argc>1 ? RXSHV_FETCH : RXSHV_NEXTV);
		NewPointer->shvname.strptr    = NULL;
		NewPointer->shvvalue.strptr   = NULL;
		NewPointer->shvnext           = NULL;            /* could be last entry! */
		if (Argc>1) {
			/* init shvname, when arguments are explicit listed */
			_REXXEXPATVERIFYALLOCATION (
			NewPointer->shvname.strptr                = (char*) REXXALLOCATEMEMORY(Argv[i].strlength + 1));
			sprintf(NewPointer->shvname.strptr, "%s\0", Argv[i].strptr);
			NewPointer->shvname.strlength             = Argv[i].strlength;
			NewPointer->shvnamelen                    = Argv[i].strlength; }
		/* execute request */
		rc = rc | RexxVariablePool(NewPointer);          /* return codes are ORed together! */
		if (Argc<=1 && (NewPointer->shvret & RXSHV_LVAR)) {
			/* halt, when last variable has been fetched */
			*SavePointerHere = NULL;
			_REXXEXPATVERIFYDEALLOCATION (
			REXXFREEMEMORY(NewPointer));            break;}
		/* concatenate list of shvblock */
		*SavePointerHere = NewPointer;
		SavePointerHere = &NewPointer->shvnext;
	}

	/* get error code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%s%d", Retstr->strptr, rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

/* ********************************************************************* *
	Parser Creation
 * ********************************************************************* *
	wrapper functions:
		APIRET APIENTRY ExpatParserCreate (APIINTERFACE);
		APIRET APIENTRY ExpatParserCreateNS (APIINTERFACE);
		APIRET APIENTRY ExpatExternalEntityParserCreate (APIINTERFACE);
		APIRET APIENTRY ExpatParserFree (APIINTERFACE);
	these functions are simply passed through to the corresponding Expat 
	functions - please refer to xmlparse.h for further descriptions!
	Furthermore they initialize (or free) the userData structure!

	general arguments:
		each function returns a reference to the created userData structure 
		except ExpatParserFree, which requires such reference as only argument
 * ********************************************************************* */

APIRET APIENTRY ExpatParserCreate (APIINTERFACE)
/*	1st argument:	encoding declaration (optional)
	2nd argument:	separation char for namespace expanded names (optional)
	returns:		a reference to userData (must not be manipulated)	*/
{
	/* local variables */
	RexxExpatData	userData = REXXALLOCATEMEMORY(sizeof(RexxExpatData));
	int				i;

	/* userData successfully allocated? */
	_REXXEXPATVERIFYALLOCATION (userData);

	/* check on valid call: up to 2 arguments */
	if (Argc == 0 || (Argc == 1 && RXNULLSTRING(Argv[0])))
		/* arguments omitted - use encoding as declared in the document */
		userData->Parser = XML_ParserCreate(NULL);
	else if ((Argc == 1 && RXVALIDSTRING(Argv[0])) ||
			 (Argc == 2 && RXVALIDSTRING(Argv[0]) && RXNULLSTRING(Argv[1])))
		/* encoding specified - override document encoding declaration! */
		userData->Parser = XML_ParserCreate(Argv[0].strptr);
	else if (Argc == 2 && RXVALIDSTRING(Argv[0]) && RXVALIDSTRING(Argv[1]))
		/* construct a parser that has namespace processing in effect	*/
		userData->Parser = XML_ParserCreateNS(Argv[0].strptr, Argv[1].strptr[0]);
	else
		return 40;									/* incorrect call	*/

	if (userData->Parser) {							/* return reference	*/
		/* init userData */
		userData->shvfirst = NULL;
		for (i = 0; i<NumberOfHandlers; i++)
			userData->Handler[i] = NULL;

		/* one userData created for each parser -> RexxExpat is thread-safe */
		XML_SetUserData(userData->Parser, userData);

		/* return string */
		sprintf(Retstr->strptr, "%p", userData);	
		Retstr->strlength = strlen(Retstr->strptr); }
	
	else {											/* creation failed	*/
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(userData));
		strcpy(Retstr->strptr, "0");	
		Retstr->strlength = strlen(Retstr->strptr); }

	return 0;
}

APIRET APIENTRY ExpatParserCreateNS (APIINTERFACE)
/* invokes ExpatParserCreate - please refere to ExpatParserCreate		*/
{
	return ExpatParserCreate (Name, Argc, Argv, Queuename, Retstr) ;
}

APIRET APIENTRY ExpatExternalEntityParserCreate (APIINTERFACE)
/*	1st argument:	reference to some userData from which new userData will be "inherited"
	2nd argument:	context argument (passed in calls to ExternalEntityRefHandler)
	3rd argument:	encoding declaration
	returns:		a reference to userData (must not be manipulated)	*/
{
	/* local variables */
	RexxExpatData	parentData, userData = REXXALLOCATEMEMORY(sizeof(RexxExpatData));
	int				i;

	/* userData successfully allocated? */
	_REXXEXPATVERIFYALLOCATION (userData);

	/* check on valid call:	must be 3 valid argument */
	if (Argc != 3 || !RXVALIDSTRING(Argv[0]) || RXNULLSTRING(Argv[1]) || RXNULLSTRING(Argv[2]))
		return 40;								/* incorrect call if not	*/

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &parentData);

	/* construct a parser for parsing external entities */
	userData->Parser = XML_ExternalEntityParserCreate(parentData->Parser, Argv[1].strptr, Argv[2].strptr);

	if (userData->Parser) {						/* return reference	*/
		/* copy userData */
		/* use same variable pool */
		userData->shvfirst = parentData->shvfirst;
//		userData->Encoding = parentData->Encoding;
		/* use own handlers - user may redefine handlers */
		for (i = 0; i < NumberOfHandlers; i++)
			_ExpatCopyHandler(parentData->Handler[i], &userData->Handler[i]);

		/* one userData created for each parser -> RexxExpat is thread-safe */
		XML_SetUserData(userData->Parser, userData);

		/* return string */
		sprintf(Retstr->strptr, "%p", userData);	
		Retstr->strlength = strlen(Retstr->strptr); }

	else {											/* creation failed	*/
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(userData));
		strcpy(Retstr->strptr, "0");	
		Retstr->strlength = strlen(Retstr->strptr); }

	return 0;
}

APIRET APIENTRY ExpatParserFree (APIINTERFACE)
/*	1st argument:	reference to userData								*/
{
	/* local variables */
	RexxExpatData	userData;
	int				i;

	/* check on valid call:	must be 1 valid argument */
	if (Argc != 1 || !(RXVALIDSTRING(Argv[0])))
		return 40;								/* incorrect call if not	*/

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* free resources */
	XML_ParserFree(userData->Parser);
	for (i = 0; i < NumberOfHandlers; i++) {
		_REXXEXPATVERIFYDEALLOCATION (
		_ExpatFreeHandler(&userData->Handler[i])); }
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(userData));

	strcpy(Retstr->strptr, "");	
	Retstr->strlength = strlen(Retstr->strptr);
	return 0;
}

/* ********************************************************************* *
	Parsing 
 * ********************************************************************* *
	wrapper functions:
		APIRET APIENTRY ExpatParse (APIINTERFACE);
	this functions is simply passed through to the corresponding Expat 
	function - please refer to xmlparse.h for further descriptions!
 * ********************************************************************* */

APIRET APIENTRY ExpatParse (APIINTERFACE)
/*	1st argument:	reference to userData								
	2nd argument:	the string to be parsed 
	3rd argument:	0, if isFinal               
	returns:		0, if an error occurred     */
{
	/* local variables */
	RexxExpatData	userData;
	void			*Buffer;

	/* check on valid call: must be 3 arguments */
	if (Argc != 3 || !RXVALIDSTRING(Argv[0]) || RXNULLSTRING(Argv[1]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* copy Argv[1].strptr */
	/* passing Argv[1].strptr directly to XML_Parse() causes memory leaks, when RexxStart's called */
	/* providing an own buffer will cause trouble as well, since the 'Parse Position Functions' are trying 
	   to access the buffer - so the buffer must stay in memory even after XML_Parse(...) has been called. 
	   So, there is no (simple) way to de-allocate the buffer :-( - but XML_ParseBuffer(...) will do :-) */
	_REXXEXPATVERIFYALLOCATION (
	Buffer = XML_GetBuffer(userData->Parser, Argv[1].strlength));
	strcpy(Buffer, Argv[1].strptr);

	/* parse */
	sprintf(Retstr->strptr, "%d", XML_ParseBuffer(userData->Parser, Argv[1].strlength, (RXNULLSTRING(Argv[2]) || RXZEROLENSTRING(Argv[2]) || !atoi(Argv[2].strptr))));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

/* ********************************************************************* *
	Handler Setting
 * ********************************************************************* *
	helper functions:
		APIRET _ExpatFreeHandler (EXPATHANDLER *Handler);
		APIRET _ExpatSetHandler  (EXPATHANDLER *Handler, RXSTRING ProcName, RXSTRING EnvName, RXSTRING Instore);
		APIRET _ExpatCopyHandler (EXPATHANDLER SourceHandler, EXPATHANDLER *DestinationHandler);
	
	wrapper functions:
		APIRET APIENTRY ExpatSetElementHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetCharacterDataHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetProcessingInstructionHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetCommentHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetDoctypeDeclHandler (APIINTERFACE)
		APIRET APIENTRY ExpatSetCdataSectionHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetDefaultHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetDefaultHandlerExpand (APIINTERFACE);
		APIRET APIENTRY ExpatSetInternalParsedEntityDeclHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetExternalParsedEntityDeclHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetExternalEntityRefHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetUnknownEncodingHandler (APIINTERFACE);
  		APIRET APIENTRY ExpatSetUnknownEncodingConverter (APIINTERFACE);
		APIRET APIENTRY ExpatSetNamespaceDeclHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetUnparsedEntityDeclHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetNotationDeclHandler (APIINTERFACE);
		APIRET APIENTRY ExpatSetNotStandaloneHandler (APIINTERFACE);
	these functions save the passed handler information to the userData structure 
	and register the appropriate handler with the Expat parser 
	except ExpatSetUnknownEncodingConverter, its handler information will be used by the UnknownEncodingHandler

	general arguments:
		for each handler the following three parameters are required
		ProcName	contains an exact image of a REXX procedure disk file, complete with carriage returns, line feeds and end-of-file characters
					=> IF isInstore is NOT 0, null, of zero length NOR left out
					or is the name of a REXX procedure disk file (optional with extension, drive, and path)
					=> IF isInstore IS 0, null, of zero length OR left out
		EnvName		specifies the initial ADDRESS environment; if left out, the file extension is used as the initial ADDRESS environment
		isInstore	decides, if ProcName is an exact image of a REXX procedure disk file or the file itself (that is its name!) 
 * ********************************************************************* */

APIRET _ExpatFreeHandler (EXPATHANDLER *Handler) 
/* frees EXPATHANDLER *Handler and all its members, if in use */
/* returns 0 if de-allocation was successful, while on error _REXXEXPATVERIFYDEALLOCATION is processed */
{
	if (*Handler) {
		/* free memory (if in use) */
		if ((*Handler)->Instore && (*Handler)->Instore[0].strptr) {
				_REXXEXPATVERIFYDEALLOCATION (	
				REXXFREEMEMORY((*Handler)->Instore[0].strptr)); }
		if ((*Handler)->Instore && (*Handler)->Instore[1].strptr) {
				_REXXEXPATVERIFYDEALLOCATION (	
				REXXFREEMEMORY((*Handler)->Instore[1].strptr)); }
		if ((*Handler)->Instore) {
				_REXXEXPATVERIFYDEALLOCATION (	
				REXXFREEMEMORY((*Handler)->Instore)); }
		if ((*Handler)->EnvName) {
				_REXXEXPATVERIFYDEALLOCATION (	
				REXXFREEMEMORY((*Handler)->EnvName)); }
		if ((*Handler)->ProcName) {
				_REXXEXPATVERIFYDEALLOCATION (	
				REXXFREEMEMORY((*Handler)->ProcName)); }
		_REXXEXPATVERIFYDEALLOCATION (	
		REXXFREEMEMORY(*Handler)); 
	}
	*Handler = NULL;
	return REXXFREEMEMORY_OK;
}

APIRET _ExpatSetHandler (EXPATHANDLER *Handler, RXSTRING ProcName, RXSTRING EnvName, RXSTRING Instore) 
/* sets EXPATHANDLER *Handler and all its members according to passed ProcName, EnvName and Instore */
/* returns 0 if allocation was successful, while on error _REXXEXPATVERIFYALLOCATION is processed */
{
	/* local variables & constants ;-) */
	#define PROCNAME_INVALID "SAY 'No valid ProcName specified!';\0"

	/* free memory (if in use) - user might redefine handlers */
	_REXXEXPATVERIFYDEALLOCATION (	
	_ExpatFreeHandler (Handler));

	/* allocate memory for handler */
	_REXXEXPATVERIFYALLOCATION (
	*Handler = REXXALLOCATEMEMORY(sizeof(EXPATHANDLER)));

	/* set members of EXPATHANDLER */
	if (RXVALIDSTRING(ProcName)) {
		if (RXNULLSTRING(Instore) || RXZEROLENSTRING(Instore) || !atoi(Instore.strptr)) {
			/* external - ProcName contains name of file with rexx code */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->ProcName	          = (char*) REXXALLOCATEMEMORY(ProcName.strlength + 1));
			sprintf((*Handler)->ProcName,    "%s\0",   ProcName.strptr);
			/* external - Instore is of no use here */
			(*Handler)->Instore = NULL; }
		else if (!strcmp(Instore.strptr, "macro") || !strcmp(Instore.strptr, "MACRO")) { 
			/* macro - ProcName contains name of rexx macro */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->ProcName	          = (char*) REXXALLOCATEMEMORY(ProcName.strlength + 1));
			sprintf((*Handler)->ProcName,    "%s\0",   ProcName.strptr);
			/* macro - both Instore.strptr are supposed to be NULL! */
			(*Handler)->Instore[0].strptr    = NULL;
			(*Handler)->Instore[1].strptr    = NULL; }
		else { 
//		if (RXVALIDSTRING(Instore)) {
			/* instore - ProcName must not be NULL */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->ProcName             = (char*) REXXALLOCATEMEMORY(1));
			sprintf((*Handler)->ProcName, "\0");       /* must not be NULL */
			/* instore - Instore[0] contains rexx code */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->Instore              = (PRXSTRING) REXXALLOCATEMEMORY(2 * sizeof (PRXSTRING)));
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->Instore[0].strptr    = (char*) REXXALLOCATEMEMORY(ProcName.strlength + 1));
			sprintf((*Handler)->Instore[0].strptr, "%s\0", ProcName.strptr);
			(*Handler)->Instore[0].strlength = strlen((*Handler)->Instore[0].strptr);
			/* instore - Instore[1] will contain translated image of Instore[0] after first call of RexxStart() */
			(*Handler)->Instore[1].strptr    = NULL;
		}
	} 
	else {	/* PROCNAME_INVALID */
			/* instore - ProcName must not be NULL */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->ProcName             = (char*) REXXALLOCATEMEMORY(1));
			sprintf((*Handler)->ProcName,"\0");        /* must not be NULL */
			/* instore - Instore[0] contains PROCNAME_INVALID */
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->Instore              = (PRXSTRING) REXXALLOCATEMEMORY(2 * sizeof (PRXSTRING)));
			_REXXEXPATVERIFYALLOCATION (
			(*Handler)->Instore[0].strptr    = (char*) REXXALLOCATEMEMORY(strlen(PROCNAME_INVALID) + 1));
			sprintf((*Handler)->Instore[0].strptr, PROCNAME_INVALID);
			(*Handler)->Instore[0].strlength = strlen((*Handler)->Instore[0].strptr);
			/* instore - Instore[1] will contain translated image of Instore[0] after first call of RexxStart() */
			(*Handler)->Instore[1].strptr    = NULL;   }

	if (RXVALIDSTRING(EnvName)) { 
		/* if EnvName is defined by user */
		_REXXEXPATVERIFYALLOCATION (
		(*Handler)->EnvName = (char*) REXXALLOCATEMEMORY(EnvName.strlength + 1));
		sprintf((*Handler)->EnvName,  "%s\0", EnvName.strptr); }
	else
		(*Handler)->EnvName = NULL;
	
	return REXXALLOCATEMEMORY_OK;
}

APIRET _ExpatCopyHandler (EXPATHANDLER SourceHandler, EXPATHANDLER *DestinationHandler) 
/* copies EXPATHANDLER SourceHandler to EXPATHANDLER *DestinationHandler */
/* returns 0 if allocation was successful, while on error _REXXEXPATVERIFYALLOCATION is processed */
{
	/* local variables */
	int i;

	/* free memory (if in use) - user might redefine handlers */
	_REXXEXPATVERIFYDEALLOCATION (	
	_ExpatFreeHandler (DestinationHandler));

	/* allocate memory for handler */
	_REXXEXPATVERIFYALLOCATION (
	*DestinationHandler = REXXALLOCATEMEMORY(sizeof(EXPATHANDLER)));

	/* copy ProcName */
	if (SourceHandler->ProcName) { 
		_REXXEXPATVERIFYALLOCATION (
		(*DestinationHandler)->ProcName = (char*) REXXALLOCATEMEMORY(strlen(SourceHandler->ProcName) + 1));
		sprintf((*DestinationHandler)->ProcName,  "%s\0", SourceHandler->ProcName); }
	else
		(*DestinationHandler)->ProcName = NULL;
	/* copy EnvName */
	if (SourceHandler->EnvName) { 
		_REXXEXPATVERIFYALLOCATION (
		(*DestinationHandler)->EnvName = (char*) REXXALLOCATEMEMORY(strlen(SourceHandler->EnvName) + 1));
		sprintf((*DestinationHandler)->EnvName,  "%s\0", SourceHandler->EnvName); }
	else
		(*DestinationHandler)->EnvName = NULL;
	/* copy Instore */
	if (SourceHandler->Instore) { 
		_REXXEXPATVERIFYALLOCATION (
		(*DestinationHandler)->Instore = (PRXSTRING) REXXALLOCATEMEMORY(2 * sizeof (PRXSTRING)));
		for (i=0; i<2; i++) {
			if (SourceHandler->Instore[i].strptr) { 
				_REXXEXPATVERIFYALLOCATION (
				(*DestinationHandler)->Instore[i].strptr    = (char*) REXXALLOCATEMEMORY(SourceHandler->Instore[i].strlength + 1));
				sprintf((*DestinationHandler)->Instore[i].strptr, "%s\0", SourceHandler->Instore[i].strptr);
				(*DestinationHandler)->Instore[i].strlength = strlen((*DestinationHandler)->Instore[i].strptr); }
			else
				(*DestinationHandler)->Instore[i].strptr = NULL; 
		} 
	}
	else
		(*DestinationHandler)->Instore = NULL;

	return 0;
}

APIRET APIENTRY ExpatSetElementHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for StartElementHandler 
	5th-7th argument:	ProcName, EnvName and isInstore for EndElementHandler  
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set StartElementHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnStartElement], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnStartElement]);

	if (!rc) {    /* if _ExpatSetHandler/_ExpatFreeHandler failed */
		/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
		sprintf(Retstr->strptr, "%d", rc);
		Retstr->strlength = strlen(Retstr->strptr);
		return 0; /* break */}

	/* set EndElementHandler */
	if (Argc >= 4 && RXVALIDSTRING(Argv[4]))
		rc = _ExpatSetHandler (&userData->Handler[OnEndElement], Argv[4], Argv[5], Argv[6]); 
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnEndElement]);

	/* register handler */
	XML_SetElementHandler(userData->Parser, 
		(userData->Handler[OnStartElement] ? _ExpatStartElementHandler : NULL), 
		(userData->Handler[OnEndElement]   ? _ExpatEndElementHandler   : NULL));

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetCharacterDataHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for CharacterDataHandler 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set CharacterDataHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnCharacterData], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnCharacterData]);

	/* register handler */
	XML_SetCharacterDataHandler(userData->Parser, 
		(userData->Handler[OnCharacterData] ? _ExpatCharacterDataHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetProcessingInstructionHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for ProcessingInstructionHandler 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set ProcessingInstructionHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnProcessingInstruction], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnProcessingInstruction]);

	/* register handler */
	XML_SetProcessingInstructionHandler(userData->Parser, 
		(userData->Handler[OnProcessingInstruction] ? _ExpatProcessingInstructionHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetCommentHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for CommentHandler 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set CommentHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnComment], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnComment]);

	/* register handler */
	XML_SetCommentHandler(userData->Parser, 
		(userData->Handler[OnComment] ? _ExpatCommentHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetDoctypeDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for StartDoctypeDeclHandler
	5th-7th argument:	ProcName, EnvName and isInstore for EndDoctypeDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set StartDoctypeDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnStartDoctypeDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnStartDoctypeDecl]);

	if (!rc) {    /* if _ExpatSetHandler/_ExpatFreeHandler failed */
		/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
		sprintf(Retstr->strptr, "%d", rc);
		Retstr->strlength = strlen(Retstr->strptr);
		return 0; /* break */}

	/* set EndDoctypeDeclHandler */
	if (Argc >= 4 && RXVALIDSTRING(Argv[4]))
		rc = _ExpatSetHandler (&userData->Handler[OnEndDoctypeDecl], Argv[4], Argv[5], Argv[6]); 
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnEndDoctypeDecl]);

	/* register handler */
	XML_SetDoctypeDeclHandler(userData->Parser, 
		(userData->Handler[OnStartDoctypeDecl] ? _ExpatStartDoctypeDeclHandler : NULL),
		(userData->Handler[OnEndDoctypeDecl]   ? _ExpatEndDoctypeDeclHandler   : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetCdataSectionHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for StartCdataSectionHandler
	5th-7th argument:	ProcName, EnvName and isInstore for EndCdataSectionHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set StartCdataSectionHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnStartCdataSection], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnStartCdataSection]);

	if (!rc) {    /* if _ExpatSetHandler/_ExpatFreeHandler failed */
		/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
		sprintf(Retstr->strptr, "%d", rc);
		Retstr->strlength = strlen(Retstr->strptr);
		return 0; /* break */}

	/* set EndCdataSectionHandler */
	if (Argc >= 4 && RXVALIDSTRING(Argv[4]))
		rc = _ExpatSetHandler (&userData->Handler[OnEndCdataSection], Argv[4], Argv[5], Argv[6]); 
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnEndCdataSection]);

	/* register handler */
	XML_SetCdataSectionHandler(userData->Parser, 
		(userData->Handler[OnStartCdataSection] ? _ExpatStartCdataSectionHandler : NULL),
		(userData->Handler[OnEndCdataSection]   ? _ExpatEndCdataSectionHandler   : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetDefaultHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for DefaultHandler 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set DefaultHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnDefault], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnDefault]);

	/* register handler */
	XML_SetDefaultHandler(userData->Parser, 
		(userData->Handler[OnDefault] ? _ExpatDefaultHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetDefaultHandlerExpand (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for DefaultHandlerExpand 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set DefaultHandlerExpand */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnDefaultExpand], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnDefaultExpand]);

	/* register handler */
	XML_SetDefaultHandlerExpand(userData->Parser,
		(userData->Handler[OnDefaultExpand] ? _ExpatDefaultHandlerExpand : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetInternalParsedEntityDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for InternalParsedEntityDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set InternalParsedEntityDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnInternalParsedEntityDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnInternalParsedEntityDecl]);

	/* register handler */
	XML_SetInternalParsedEntityDeclHandler(userData->Parser, 
		(userData->Handler[OnInternalParsedEntityDecl] ? _ExpatInternalParsedEntityDeclHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetExternalParsedEntityDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for ExternalParsedEntityDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set ExternalParsedEntityDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnExternalParsedEntityDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnExternalParsedEntityDecl]);

	/* register handler */
	XML_SetExternalParsedEntityDeclHandler(userData->Parser, 
		(userData->Handler[OnExternalParsedEntityDecl] ? _ExpatExternalParsedEntityDeclHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetExternalEntityRefHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for ExternalEntityRefHandler 
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set ExternalEntityRefHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnExternalEntityRef], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnExternalEntityRef]);

	/* register handler */
	XML_SetExternalEntityRefHandler(userData->Parser, 
		(userData->Handler[OnExternalEntityRef] ? _ExpatExternalEntityRefHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetUnknownEncodingHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for UnknownEncodingHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set UnknownEncodingHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnUnknownEncoding], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnUnknownEncoding]);

	/* register handler */
	XML_SetUnknownEncodingHandler(userData->Parser, 
		(userData->Handler[OnUnknownEncoding] ? _ExpatUnknownEncodingHandler : NULL),
		userData); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetUnknownEncodingConverter (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for UnknownEncodingConvertFunction
	5th-7th argument:	ProcName, EnvName and isInstore for UnknownEncodingReleaseFunction
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set UnknownEncodingConvertFunction */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnUnknownEncodingConvert], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnUnknownEncodingConvert]);

	if (!rc) {    /* if _ExpatSetHandler/_ExpatFreeHandler failed */
		/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
		sprintf(Retstr->strptr, "%d", rc);
		Retstr->strlength = strlen(Retstr->strptr);
		return 0; /* break */}

	/* set UnknownEncodingReleaseFunction */
	if (Argc >= 4 && RXVALIDSTRING(Argv[4]))
		rc = _ExpatSetHandler (&userData->Handler[OnUnknownEncodingRelease], Argv[4], Argv[5], Argv[6]); 
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnUnknownEncodingRelease]);

	/* register handler */
	/* this is done in the UnknownEncodingHandler ! */

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetNamespaceDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for StartNamespaceDeclHandler
	5th-7th argument:	ProcName, EnvName and isInstore for EndNamespaceDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set StartNamespaceDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnStartNamespaceDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnStartNamespaceDecl]);

	if (!rc) {    /* if _ExpatSetHandler/_ExpatFreeHandler failed */
		/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
		sprintf(Retstr->strptr, "%d", rc);
		Retstr->strlength = strlen(Retstr->strptr);
		return 0; /* break */}

	/* set EndNamespaceDeclHandler */
	if (Argc >= 4 && RXVALIDSTRING(Argv[4]))
		rc = _ExpatSetHandler (&userData->Handler[OnEndNamespaceDecl], Argv[4], Argv[5], Argv[6]); 
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnEndNamespaceDecl]);

	/* register handler */
	XML_SetNamespaceDeclHandler(userData->Parser, 
		(userData->Handler[OnStartNamespaceDecl] ? _ExpatStartNamespaceDeclHandler : NULL),
		(userData->Handler[OnEndNamespaceDecl]   ? _ExpatEndNamespaceDeclHandler   : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetUnparsedEntityDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for UnparsedEntityDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set UnparsedEntityDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnUnparsedEntityDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnUnparsedEntityDecl]);

	/* register handler */
	XML_SetUnparsedEntityDeclHandler(userData->Parser, 
		(userData->Handler[OnUnparsedEntityDecl] ? _ExpatUnparsedEntityDeclHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetNotationDeclHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for NotationDeclHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set NotationDeclHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnNotationDecl], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnNotationDecl]);

	/* register handler */
	XML_SetNotationDeclHandler(userData->Parser, 
		(userData->Handler[OnNotationDecl] ? _ExpatNotationDeclHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetNotStandaloneHandler (APIINTERFACE)
/*	1st argument:		reference to userData
	2nd-4th argument:	ProcName, EnvName and isInstore for NotStandaloneHandler
*/
{
	/* local variables */
	RexxExpatData	userData;
	int				rc = 0;

	/* check on valid call: must be 1 valid argument */
	if (Argc < 1)
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* set NotStandaloneHandler */
	if (Argc >= 1 && RXVALIDSTRING(Argv[1]))
		rc = _ExpatSetHandler (&userData->Handler[OnNotStandalone], Argv[1], Argv[2], Argv[3]);
	else
		rc = _ExpatFreeHandler(&userData->Handler[OnNotStandalone]);

	/* register handler */
	XML_SetNotStandaloneHandler(userData->Parser, 
		(userData->Handler[OnNotStandalone] ? _ExpatNotStandaloneHandler : NULL)); 

	/* set return code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf(Retstr->strptr, "%d", rc);
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

/* ********************************************************************* *
	Handlers - RexxExpat Handlers
 * ********************************************************************* *
	helper functions:
		APIRET _ExpatMakeArgList    (ULONG ArgCount, PRXSTRING *DestinationList, char **SourceList);
		APIRET _ExpatFreeArgList    (ULONG ArgCount, PRXSTRING *ArgList);
		APIRET _ExpatInsert2ArgList (PRXSTRING *DestinationList, int EntryIndex, const char *ListEntry, int EntryLength);

	wrapper functions:
	RexxExpat handler functions registered with Expat (XML_Set...Handler)
		void _ExpatStartElementHandler             (void *pUserData, const XML_Char *name, const XML_Char **atts);
		void _ExpatEndElementHandler               (void *pUserData, const XML_Char *name);
		void _ExpatStartNamespaceDeclHandler       (void *pUserData, const XML_Char *prefix, const XML_Char *uri);
		void _ExpatEndNamespaceDeclHandler         (void *pUserData, const XML_Char *prefix);
		void _ExpatProcessingInstructionHandler    (void *pUserData, const XML_Char *target, const XML_Char *data);
		void _ExpatCommentHandler                  (void *pUserData, const XML_Char *data);
		void _ExpatNotationDeclHandler             (void *pUserData, const XML_Char *notationName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
		void _ExpatUnparsedEntityDeclHandler       (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId, const XML_Char *notationName);
		void _ExpatInternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *replacementText, int replacementTextLength);
		void _ExpatExternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
		int  _ExpatExternalEntityRefHandler        (XML_Parser parser, const XML_Char *context, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
		void _ExpatDefaultHandler                  (void *pUserData, const XML_Char *s, int len);
		void _ExpatDefaultHandlerExpand            (void *pUserData, const XML_Char *s, int len);
		void _ExpatCharacterDataHandler            (void *pUserData, const XML_Char *s, int len);
		void _ExpatStartDoctypeDeclHandler         (void *pUserData, const XML_Char *doctypeName);
		void _ExpatEndDoctypeDeclHandler           (void *pUserData);
		void _ExpatStartCdataSectionHandler        (void *pUserData);
		void _ExpatEndCdataSectionHandler          (void *pUserData);
		int  _ExpatNotStandaloneHandler            (void *pUserData);
		int  _ExpatUnknownEncodingHandler          (void *pUserData, const XML_Char *name, XML_Encoding *info);
	RexxExpat encoding conversion functions pointed to in XML_Encoding structure
		int  _ExpatUnknownEncodingConvert          (void *pUserData, const char *s);
		void _ExpatUnknownEncodingRelease          (void *pUserData);
	these functions are simply passed through from the corresponding Expat 
	functions - please refer to xmlparse.h for further descriptions!
 * ********************************************************************* */

APIRET _ExpatFreeArgList (ULONG ArgCount, PRXSTRING *ArgList) 
/* frees PRXSTRING ArgList and all its members, if in use */
{

	/* local variables */
	ULONG i;

	/* free memory (if in use) */
	if (*ArgList) {
		for (i=0; i<ArgCount; i++) {
			if ((*ArgList)[i].strptr) {
				_REXXEXPATVERIFYDEALLOCATION (
				REXXFREEMEMORY((*ArgList)[i].strptr)); }
		}
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(*ArgList)); }
	*ArgList = NULL;

	return REXXFREEMEMORY_OK;
}

APIRET _ExpatMakeArgList (ULONG ArgCount, PRXSTRING *DestinationList, char **SourceList)
/* copies the array of char* 'SourceList' to the array of RXSTRINGs 'DestinationList' */
{
	/* local variables */
	ULONG i;

	/* free memory (if in use) */
	_REXXEXPATVERIFYDEALLOCATION (	
	_ExpatFreeArgList (ArgCount, DestinationList));

	/* set ArgList  */
	_REXXEXPATVERIFYALLOCATION (
	*DestinationList = REXXALLOCATEMEMORY(ArgCount * sizeof(RXSTRING)));
	for (i=0; i<ArgCount; i++) {
		if (SourceList[i]) {     /* SourceList[i] is not NULL */
			_REXXEXPATVERIFYALLOCATION (
			(*DestinationList)[i].strptr = (char*) REXXALLOCATEMEMORY(strlen(SourceList[i])));
			strcpy((*DestinationList)[i].strptr, SourceList[i]);
			(*DestinationList)[i].strlength = strlen((*DestinationList)[i].strptr); }
		else {                   /* SourceList[i] is     NULL */
			(*DestinationList)[i].strptr = NULL;
			(*DestinationList)[i].strlength = 0; }
	}

	return REXXALLOCATEMEMORY_OK;
}

APIRET _ExpatInsert2ArgList (PRXSTRING *DestinationList, int EntryIndex, const char *ListEntry, int EntryLength)
/* inserts const char* 'ListEntry' to array of RXSTRINGs 'DestinationList' at position 'EntryIndex' */
/* 'EntryIndex' must be a valid index of 'DestinationList'... (use _ExpatMakeArgList(...) with 'DestinationList[EntryIndex]' set to NULL) */
{
	/* no local variables needed */

	/* *DestinationList must be allocated already */
	if (!*DestinationList) return REXXALLOCATEMEMORY_FAILED;

	/* free memory (if in use) */
	if ((*DestinationList)[EntryIndex].strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY((*DestinationList)[EntryIndex].strptr)); }

	/* insert to ArgList  */
	_REXXEXPATVERIFYALLOCATION (
	(*DestinationList)[EntryIndex].strptr = (char*) REXXALLOCATEMEMORY(EntryLength));
	strcpy((*DestinationList)[EntryIndex].strptr, ListEntry);
	(*DestinationList)[EntryIndex].strlength = EntryLength;

	return REXXALLOCATEMEMORY_OK;
}

void _ExpatStartElementHandler (void *pUserData, const XML_Char *name, const XML_Char **atts)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char**			args;
	long			ArgCount;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				i, j, rc, currentExpatHandlerType=OnStartElement;
	
	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* get number of atts - two additional args needed for name and parser */
	for (ArgCount=0; atts[ArgCount]; ArgCount++); ArgCount+=2;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args    = (char**) REXXALLOCATEMEMORY(sizeof(char*) * ArgCount));
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*)  REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) name;
	for (i=2, j=0; atts[j]; i++, j++)
		args[i] = (char*) atts[j];
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* if anything's returned, pass markup to default handler */
	if (RXVALIDSTRING(result_str))     
		XML_DefaultCurrent (userData->Parser); 

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatEndElementHandler (void *pUserData, const XML_Char *name)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnEndElement;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) name;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* if anything's returned, pass markup to default handler */
	if (RXVALIDSTRING(result_str))     
		XML_DefaultCurrent (userData->Parser); 

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatStartNamespaceDeclHandler (void *pUserData, const XML_Char *prefix, const XML_Char *uri)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [3];
	const ULONG		ArgCount = 3;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnStartNamespaceDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) prefix;
	args[2] = (char*) uri;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatEndNamespaceDeclHandler (void *pUserData, const XML_Char *prefix) 
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnEndNamespaceDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) prefix;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatProcessingInstructionHandler (void *pUserData, const XML_Char *target, const XML_Char *data)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [3];
	const ULONG		ArgCount = 3;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnProcessingInstruction;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) target;
	args[2] = (char*) data;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* if anything's returned, pass markup to default handler */
	if (RXVALIDSTRING(result_str))     
		XML_DefaultCurrent (userData->Parser); 

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatCommentHandler (void *pUserData, const XML_Char *data)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnComment;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) data;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatNotationDeclHandler (void *pUserData, const XML_Char *notationName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [5];
	const ULONG		ArgCount = 5;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnNotationDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) notationName;
	args[2] = (char*) base;
	args[3] = (char*) systemId;
	args[4] = (char*) publicId;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatUnparsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId, const XML_Char *notationName)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [6];
	const ULONG		ArgCount = 6;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnUnparsedEntityDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) entityName;
	args[2] = (char*) base;
	args[3] = (char*) systemId;
	args[4] = (char*) publicId;
	args[5] = (char*) notationName;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatInternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *replacementText, int replacementTextLength)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [3];
	const ULONG		ArgCount = 3;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnInternalParsedEntityDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) entityName;
	args[2] = NULL; /* to be set by _ExpatInsert2ArgList(...) */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYALLOCATION (
	_ExpatInsert2ArgList (&ArgList, 2, replacementText, replacementTextLength));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatExternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [5];
	const ULONG		ArgCount = 5;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnExternalParsedEntityDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) entityName;
	args[2] = (char*) base;
	args[3] = (char*) systemId;
	args[4] = (char*) publicId;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

int  _ExpatExternalEntityRefHandler (XML_Parser parser, const XML_Char *context, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId) 
{
	/* local variables - note: userData is retrieved by XML_GetUserData() */
	RexxExpatData	userData = (RexxExpatData) 	XML_GetUserData(parser);
	char*			args	  [5];
	const ULONG		ArgCount = 5;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnExternalEntityRef;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return 0;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);
	args[1] = (char*) context;
	args[2] = (char*) base;
	args[3] = (char*) systemId;
	args[4] = (char*) publicId;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;

	return result_int;
}

void _ExpatDefaultHandler (void *pUserData, const XML_Char *s, int len)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnDefault;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = NULL;	/* to be set by _ExpatInsert2ArgList(...) */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYALLOCATION (
	_ExpatInsert2ArgList (&ArgList, 1, s, len));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatDefaultHandlerExpand (void *pUserData, const XML_Char *s, int len)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnDefaultExpand;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = NULL;	/* to be set by _ExpatInsert2ArgList(...) */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYALLOCATION (
	_ExpatInsert2ArgList (&ArgList, 1, s, len));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)) };
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatCharacterDataHandler (void *pUserData, const XML_Char *s, int len)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnCharacterData;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = NULL;	/* to be set by _ExpatInsert2ArgList(...) */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYALLOCATION (
	_ExpatInsert2ArgList (&ArgList, 1, s, len));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* if anything's returned, pass markup to default handler */
	if (RXVALIDSTRING(result_str))     
		XML_DefaultCurrent (userData->Parser); 

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatStartDoctypeDeclHandler (void *pUserData, const XML_Char *doctypeName)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnStartDoctypeDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) doctypeName;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatEndDoctypeDeclHandler (void *pUserData) 
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [1];
	const ULONG		ArgCount = 1;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnEndDoctypeDecl;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatStartCdataSectionHandler (void *pUserData)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [1];
	const ULONG		ArgCount = 1;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnStartCdataSection;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

void _ExpatEndCdataSectionHandler (void *pUserData)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [1];
	const ULONG		ArgCount = 1;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnEndCdataSection;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

int  _ExpatNotStandaloneHandler (void *pUserData)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [1];
	const ULONG		ArgCount = 1;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnNotStandalone;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return 0;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;

	return result_int;
}

int  _ExpatUnknownEncodingHandler (void *pUserData, const XML_Char *name, XML_Encoding *info) 
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char			*args	  [2], *currIndex;
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				i, rc, currentExpatHandlerType=OnUnknownEncoding;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return 0;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = (char*) name;
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* set XML_Encoding structure */
	currIndex = result_str.strptr;
	for (i=0; currIndex && i<ENCODINGMAPSIZE && currIndex<=(result_str.strptr + result_str.strlength); i++, currIndex++) {
		if (isdigit((unsigned char)*currIndex)) /* this byte's defined  */
			info->map[i] = strtol(currIndex, &currIndex, 0);
		else                                    /* this byte's left out */
			info->map[i] = i;
	}
	for (; i<ENCODINGMAPSIZE; i++)              /* fill up bytes */
		info->map[i] = i;
	info->data    = (void *) userData;
	info->convert = (userData->Handler[OnUnknownEncodingConvert] ? _ExpatUnknownEncodingConvert : NULL); 
	info->release = (userData->Handler[OnUnknownEncodingRelease] ? _ExpatUnknownEncodingRelease : NULL); 

	/* remember map in userData->Encoding, because we need to know the length of the sequence of bytes in _ExpatUnknownEncodingConvert */
	for (i=0; i<ENCODINGMAPSIZE; i++)
		userData->Encoding.map[i] = info->map[i];

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;

	return (RXVALIDSTRING(result_str));
}

int  _ExpatUnknownEncodingConvert (void *pUserData, const char *s)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [2];
	const ULONG		ArgCount = 2;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnUnknownEncodingConvert;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return -1;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	args[1] = NULL; /* to be set by _ExpatInsert2ArgList(...) */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	/* s points to a n-byte sequence where map[(unsigned char)*s] == -n and is not 0 terminated */
	_REXXEXPATVERIFYALLOCATION (
	_ExpatInsert2ArgList (&ArgList, 1, s, userData->Encoding.map[(unsigned char)*s] * (-1)));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;

	return result_int;
}

void _ExpatUnknownEncodingRelease (void *pUserData)
{
	/* local variables */
	RexxExpatData	userData = (RexxExpatData)	pUserData;
	char*			args	  [1];
	const ULONG		ArgCount = 1;
	PRXSTRING		ArgList = NULL;
	short			result_int;
	RXSTRING		result_str;
	int				rc, currentExpatHandlerType=OnUnknownEncodingRelease;

	/* break - if handler's not set */
	if (!userData->Handler[currentExpatHandlerType]) return;

	/* make ArgList */
	_REXXEXPATVERIFYALLOCATION (
	args[0] = (char*) REXXALLOCATEMEMORY(sizeof(userData)));
	sprintf(args[0], "%p", userData);	
	_REXXEXPATVERIFYALLOCATION (
	_ExpatMakeArgList (ArgCount, &ArgList, args));
	_REXXEXPATVERIFYDEALLOCATION (
	REXXFREEMEMORY(args[0]));

	/* allocate result_str */
	/* in fact, RexxStart is supposed to do this, but obviously it doesn't... */
	_REXXEXPATVERIFYALLOCATION (
	result_str.strptr = (char*) REXXALLOCATEMEMORY(256));

	_REXXEXPATVERIFYREXXSTART (
	/* call RexxStart */
	rc = RexxStart (
		ArgCount,												/* number of arguments                  */
		ArgList,												/* pointer to argument array            */
		userData->Handler[currentExpatHandlerType]->ProcName,	/* REXX procedure (code or file name)   */
		userData->Handler[currentExpatHandlerType]->Instore,	/* is ProcName code or file name?       */
		userData->Handler[currentExpatHandlerType]->EnvName,	/* initial ADDRESS environment          */
		RXSUBROUTINE,											/* call mode (here always RXSUBROUTINE) */
		NULL,													/* RexxRegisterExitDll not implemented! */
		&result_int,											/* returned result as int               */
		&result_str));											/* returned result as RXSTRING          */

	/* free memory */
	if (result_str.strptr) {
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(result_str.strptr)); }
	_REXXEXPATVERIFYDEALLOCATION (
	_ExpatFreeArgList (ArgCount, &ArgList));

	/* wait for termination of all asynchronious rexx procedures */
	REXXWAITFORTERMINATION;
}

/* ********************************************************************* *
	Parse Position and Error Reporting Functions
 * ********************************************************************* *
	wrapper functions:
		APIRET APIENTRY ExpatGetErrorCode           (APIINTERFACE);
		APIRET APIENTRY ExpatErrorString            (APIINTERFACE);
		APIRET APIENTRY ExpatGetCurrentByteIndex    (APIINTERFACE);
		APIRET APIENTRY ExpatGetCurrentByteCount    (APIINTERFACE);
		APIRET APIENTRY ExpatGetCurrentLineNumber   (APIINTERFACE);
		APIRET APIENTRY ExpatGetCurrentColumnNumber (APIINTERFACE);
	these functions are simply passed through to the corresponding Expat 
	functions - please refer to xmlparse.h for further descriptions!

	general arguments:
		each function requires a reference to userData as only argument
		expect ExpatErrorString, which expects an error code instead!

	these functions will break on incorrect calls only
	except ExpatErrorString, which will break on problems with allocating or de-allocation memory as well
 * ********************************************************************* */

APIRET APIENTRY ExpatGetErrorCode (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		returns type of error         */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* get error code - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetErrorCode(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatErrorString (APIINTERFACE)
/*	1st argument:	an error code 
	returns:		returns a string describing the error corresponding to the passed code */
{
	/* local variables */
	const char	*ErrorString;

	/* check on valid call: must be 1 non-empty argument */
	if (Argc != 1 || RXNULLSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get error string */
	/* XML_ErrorString may return NULL, if error code is 0 (invalid?) */
	ErrorString = XML_ErrorString(atoi(Argv[0].strptr));

	if (ErrorString && (strlen(ErrorString) + 1 > RXAUTOBUFLEN)) {
		/* define new Retstr-strptr of sufficient length */
		_REXXEXPATVERIFYDEALLOCATION (
		REXXFREEMEMORY(Retstr->strptr));
		_REXXEXPATVERIFYALLOCATION (
		Retstr->strptr = REXXALLOCATEMEMORY(strlen(ErrorString) + 1)); }

	/* set return string */
	sprintf(Retstr->strptr, "%s\0", ErrorString);
	Retstr->strlength = strlen(Retstr->strptr); 

	return 0;
}

APIRET APIENTRY ExpatGetCurrentByteIndex (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		byte offset of the position          */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* get byte index - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetCurrentByteIndex(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetCurrentByteCount (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		the number of bytes in the current event */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* get base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetCurrentByteCount(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetCurrentLineNumber (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		line number of the position           */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* get line number - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetCurrentLineNumber(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetCurrentColumnNumber (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		offset of the position - relative to the beginning of the current line */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* get column number - must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetCurrentColumnNumber(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

/* ********************************************************************* *
	Miscellaneous Functions
 * ********************************************************************* *
	wrapper functions:
		APIRET APIENTRY ExpatSetBase                    (APIINTERFACE);
		APIRET APIENTRY ExpatGetBase                    (APIINTERFACE);
		APIRET APIENTRY ExpatGetSpecifiedAttributeCount (APIINTERFACE);
		APIRET APIENTRY ExpatGetIdAttributeIndex        (APIINTERFACE);
		APIRET APIENTRY ExpatSetEncoding                (APIINTERFACE);
		APIRET APIENTRY ExpatSetParamEntityParsing      (APIINTERFACE);
	these functions are simply passed through to the corresponding Expat 
	functions - please refer to xmlparse.h for further descriptions!

	general arguments:
		each function requires a reference to userData as first argument
		the 'Set...' functions expect one further argument with the value to be set

	these functions will break on incorrect calls only
	except ExpatGetBase, which will break on problems with allocating or de-allocation memory as well
 * ********************************************************************* */

APIRET APIENTRY ExpatSetBase (APIINTERFACE)
/*	1st argument:	reference to userData 
	2nd argument:	base to be used for resolving relative URIs in system identifiers 
	returns:		0, if memory insufficient     */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 2 arguments */
	if (Argc != 2 || !RXVALIDSTRING(Argv[0]) || RXNULLSTRING(Argv[1]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* set base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_SetBase(userData->Parser, Argv[1].strptr));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetBase (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		returns the base for resolving relative URIs in system identifiers */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* define new Retstr-strptr of sufficient length */
	if (REXXFREEMEMORY(Retstr->strptr)) return REXXFREEMEMORY_FAILED;
	Retstr->strptr = REXXALLOCATEMEMORY(strlen(XML_GetBase(userData->Parser)) + 1);
	if (!Retstr->strptr) return REXXALLOCATEMEMORY_FAILED;
	
	/* get base */
	sprintf(Retstr->strptr, "%s\0", XML_GetBase(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetSpecifiedAttributeCount (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		number of explicit specified attributes in the **atts-vector of the last call to StartElementHandler */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* get base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetSpecifiedAttributeCount(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatGetIdAttributeIndex (APIINTERFACE)
/*	1st argument:	reference to userData 
	returns:		index of the ID attribute passed in the last call to StartElementHandler */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 1 valid argument */
	if (Argc != 1 || !RXVALIDSTRING(Argv[0]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);
	
	/* get base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_GetIdAttributeIndex(userData->Parser));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetEncoding (APIINTERFACE)
/*	1st argument:	reference to userData 
	2nd argument:	encoding declaration  
	returns:		not defined                    */
{
	/* local variables */
	RexxExpatData	userData;

	/* check on valid call: must be 2 arguments */
	if (Argc != 2 || !RXVALIDSTRING(Argv[0]) || RXNULLSTRING(Argv[1]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* get base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
	sprintf (Retstr->strptr, "%d", XML_SetEncoding(userData->Parser, Argv[1].strptr));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

APIRET APIENTRY ExpatSetParamEntityParsing (APIINTERFACE)
/*	1st argument:	reference to userData 
	2nd argument:	a valid code from enum xmlparse::XML_ParamEntityParsing
	returns:		0, if parameter entity parsing is not in effect */
{
	/* local variables */
	RexxExpatData	userData;
	int				code;

	/* check on valid call: must be 2 valid arguments */
	if (Argc != 2 || !RXVALIDSTRING(Argv[0]) || !RXVALIDSTRING(Argv[1]))
		return 40;								/* incorrect call if not */

	/* get reference to user data */
	sscanf (Argv[0].strptr, "%p", &userData);

	/* if user used entries of enum XML_ParamEntityParsing */
	if (!strcmp(Argv[1].strptr, "XML_PARAM_ENTITY_PARSING_NEVER"))
		code = XML_PARAM_ENTITY_PARSING_NEVER;
	else if (!strcmp(Argv[1].strptr, "XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE"))
		code = XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE;
	else if (!strcmp(Argv[1].strptr, "XML_PARAM_ENTITY_PARSING_ALWAYS"))
		code = XML_PARAM_ENTITY_PARSING_ALWAYS;
	else
		code = atoi(Argv[1].strptr);

	/* get base - rc must not be greater than 10^RXAUTOBUFLEN ;-) */
//	sprintf (Retstr->strptr, "%d", XML_SetParamEntityParsing(userData->Parser, (code>=0 ? code : atoi(Argv[1].strptr))));
	sprintf (Retstr->strptr, "%d", XML_SetParamEntityParsing(userData->Parser, code));
	Retstr->strlength = strlen(Retstr->strptr);

	return 0;
}

/* bye ;-) */