/*
 *  RexxExpat - Version 1.0 - 15th January, 2001
 *  Copyright (C) 2001 Dominik Stein <dominik_stein@gmx.de>
 *
 *  This library wraps the C-XML-parser 'Expat' so it can be used 
 *  from a Rexx environment. Since the wrapper functions mostly simply
 *  pass through the values from Expat to Rexx and vice versa, please
 *  have a look at the following ressources for further readings:
 *
 *  for documentation of the expat-interface please refer to
 *   * xmlparse.h
 *   * Clark Cooper, Using Expat, Sept. 1, 1999, XML.com, O'Reilly & Associates (http://www.xml.com/pub/1999/09/expat/reference.html)
 *
 *  for documentaiton of the Rexx SAA API please refere to
 *   * rexxsaa.h
 *   * IBM, Object REXX (for Windows NT and Windows95) Programming Guide (Version 1.03), third edition, May 1999 (<ftp://service.boulder.ibm.com/ps/products/ad/obj-xx/rexxpg.zip)>)
 *   * Anders Christensen, The Regina Rexx Interpreter, August 14, 2000 (<ftp://ftp.lightlink.com/pub/hessling/Regina/reginapdf22.zip>) 
 *
 *  for matters specific to RexxExpat please read the supplemented
 *   * reference.html
 * 
 */

/* ********************************************************************* *
	macro declarations
 * ********************************************************************* */

#ifndef RexxExpat_INCLUDED
#define RexxExpat_INCLUDED 1

#ifndef REXXEXPATAPI
#define REXXEXPATAPI 
#endif

/* Who Am I? */
#define DLLNAME "rexxexpat"

/* Interface Settings (define RexxSAA API to external functions) */
//#define APIINTERFACE PSZ Name, ULONG Argc, RXSTRING Argv[], PSZ Queuename, PRXSTRING Retstr 
/* this is according to IBM's Programming Guide */
#define APIINTERFACE PCSZ Name, ULONG Argc, PRXSTRING Argv, PCSZ Queuename, PRXSTRING Retstr 
/* this is according to Reginas rexxsaa.h */

/* Constant Settings */
/* size of map array in XML_Encoding */
#define ENCODINGMAPSIZE				256	/* byte range: 00-FF */
#define RXSHV_FATAL					127	

/* Memory Allocation Settings */
/* define method to allocate memory   - Reginas rexxsaa.h has this already defined */
/* REXXALLOCATEMEMORY must return (*void) if successful, NULL if allocation failed */
#ifndef REXXALLOCATEMEMORY
#define REXXALLOCATEMEMORY GlobalAlloc
#endif
/* define method to de-allocate memory - Reginas rexxsaa.h has this already defined */
/* REXXFREEMEMORY must return NULL if successful, anything else if de-allocation failed */
#ifndef REXXFREEMEMORY
#define REXXFREEMEMORY GlobalFree
#endif

/* Action Settings (e.g. for error handling) */
/* define action, if allocation of memory failed */
#define REXXALLOCATEMEMORY_OK		  3
#define REXXALLOCATEMEMORY_FAILED	  0
#define _REXXEXPATVERIFYALLOCATION(x)   if (!(x))  {fprintf (stderr, "RexxExpat: insufficient memory\n"); exit (REXXALLOCATEMEMORY_FAILED);}
/* define action, if de-allocation of memory failed */
#define REXXFREEMEMORY_OK			  0
#define REXXFREEMEMORY_FAILED		  5
#define _REXXEXPATVERIFYDEALLOCATION(x) if ( (x))  {fprintf (stderr, "RexxExpat: de-allocating memory failed\n"); }
/* define action, if RexxStart returns error */
#define _REXXEXPATVERIFYREXXSTART(x)    if ((x)!=0) fprintf (stderr, "RexxExpat (Error %d): %s\n", rc, (rc<0 ? "rexx interpreter error" : "can't find/load Interpreter" ));
/* define action, if asynchronious Rexx procedures are supported */
#define REXXWAITFORTERMINATION




#ifdef __cplusplus
extern "C" {
#endif

/* ********************************************************************* *
	functions to be exported
 * ********************************************************************* */

/* Load/Unload RexxExpat */
RexxFunctionHandler ExpatLoadFuncs;
RexxFunctionHandler ExpatDropFuncs;

/* Variable Pool Function */
RexxFunctionHandler ExpatSetVariablePool;
RexxFunctionHandler ExpatGetVariablePool;
RexxFunctionHandler ExpatFreeVariablePool;

/* Parser Creation */
RexxFunctionHandler ExpatParserCreate;
RexxFunctionHandler ExpatParserCreateNS;
RexxFunctionHandler ExpatExternalEntityParserCreate;
RexxFunctionHandler ExpatParserFree;

/* Parsing */
RexxFunctionHandler ExpatParse;

/* Handler Setting */
RexxFunctionHandler ExpatSetElementHandler;
RexxFunctionHandler ExpatSetCharacterDataHandler;
RexxFunctionHandler ExpatSetProcessingInstructionHandler;
RexxFunctionHandler ExpatSetCommentHandler;
RexxFunctionHandler ExpatSetDoctypeDeclHandler;
RexxFunctionHandler ExpatSetCdataSectionHandler;
RexxFunctionHandler ExpatSetDefaultHandler;
RexxFunctionHandler ExpatSetDefaultHandlerExpand;
RexxFunctionHandler ExpatSetInternalParsedEntityDeclHandler;
RexxFunctionHandler ExpatSetExternalParsedEntityDeclHandler;
RexxFunctionHandler ExpatSetExternalEntityRefHandler;
RexxFunctionHandler ExpatSetUnknownEncodingHandler;
RexxFunctionHandler ExpatSetUnknownEncodingConverter;
RexxFunctionHandler ExpatSetNamespaceDeclHandler;
RexxFunctionHandler ExpatSetUnparsedEntityDeclHandler;
RexxFunctionHandler ExpatSetNotationDeclHandler;
RexxFunctionHandler ExpatSetNotStandaloneHandler;

/* Parse Position and Error Reporting Functions */
RexxFunctionHandler ExpatGetErrorCode;
RexxFunctionHandler ExpatErrorString;
RexxFunctionHandler ExpatGetCurrentByteIndex;
RexxFunctionHandler ExpatGetCurrentByteCount;
RexxFunctionHandler ExpatGetCurrentLineNumber;
RexxFunctionHandler ExpatGetCurrentColumnNumber;

/* Miscellaneous Functions */
RexxFunctionHandler ExpatSetBase;
RexxFunctionHandler ExpatGetBase;
RexxFunctionHandler ExpatGetSpecifiedAttributeCount;
RexxFunctionHandler ExpatGetIdAttributeIndex;
RexxFunctionHandler ExpatSetEncoding;
RexxFunctionHandler ExpatSetParamEntityParsing;

/* ********************************************************************* *
    data structures for intern use *
 * ********************************************************************* */

/* ExpatHandlers - how many are there? */
enum EXPATHANDLERTYPES {
		OnStartElement, 
		OnEndElement, 
		OnCharacterData,
		OnProcessingInstruction,
		OnComment,
		OnStartDoctypeDecl,
		OnEndDoctypeDecl,
		OnStartCdataSection,
		OnEndCdataSection,
		OnDefault,
		OnDefaultExpand,
		OnExternalEntityRef,
		OnUnknownEncoding,
		OnUnknownEncodingConvert,
		OnUnknownEncodingRelease,
		OnStartNamespaceDecl,
		OnEndNamespaceDecl,
		OnUnparsedEntityDecl,
		OnInternalParsedEntityDecl,
		OnExternalParsedEntityDecl,
		OnNotationDecl,
		OnNotStandalone,
		NumberOfHandlers};

/* ExpatHandler - data storage */
typedef struct TypeDefExpatHandler {
	PSZ			ProcName;
	PSZ			EnvName;
	PRXSTRING	Instore;	/* pointer to array[2] */
} *EXPATHANDLER;

/* RexxExpatData - to make it thread-safe */
typedef struct TypeDefRexxExpatData {
	/* Expat-Data */	
	XML_Parser		Parser;                    /* parser   */
	XML_Encoding	Encoding;                  /* encoding */
	EXPATHANDLER	Handler[NumberOfHandlers]; /* handler  */
	/* Rexx-Data */
	PSHVBLOCK		shvfirst;        /* rexx variable pool */
} *RexxExpatData;

/* ********************************************************************* *
	functions for intern use
 * ********************************************************************* */

/* helper functions used by variable pool functions */
APIRET _ExpatFreeRexxVariablePool (PSHVBLOCK *FirstPointer);

/* helper functions used by handler setting procs */
APIRET _ExpatFreeHandler (EXPATHANDLER *Handler);
APIRET _ExpatSetHandler  (EXPATHANDLER *Handler, RXSTRING ProcName, RXSTRING EnvName, RXSTRING Instore);
APIRET _ExpatCopyHandler (EXPATHANDLER SourceHandler, EXPATHANDLER *DestinationHandler);

/* helper functions used in ExpatHandlers */
APIRET _ExpatMakeArgList    (ULONG ArgCount, PRXSTRING *DestinationList, char **SourceList);
APIRET _ExpatFreeArgList    (ULONG ArgCount, PRXSTRING *ArgList);
APIRET _ExpatInsert2ArgList (PRXSTRING *DestinationList, int EntryIndex, const char *ListEntry, int EntryLength);

/* ExpatHandlers - here they come */
void _ExpatStartElementHandler             (void *pUserData, const XML_Char *name, const XML_Char **atts);
void _ExpatEndElementHandler               (void *pUserData, const XML_Char *name);
void _ExpatStartNamespaceDeclHandler       (void *pUserData, const XML_Char *prefix, const XML_Char *uri);
void _ExpatEndNamespaceDeclHandler         (void *pUserData, const XML_Char *prefix);
void _ExpatProcessingInstructionHandler    (void *pUserData, const XML_Char *target, const XML_Char *data);
void _ExpatCommentHandler                  (void *pUserData, const XML_Char *data);
void _ExpatNotationDeclHandler             (void *pUserData, const XML_Char *notationName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
void _ExpatUnparsedEntityDeclHandler       (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId, const XML_Char *notationName);
void _ExpatInternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *replacementText, int replacementTextLength);
void _ExpatExternalParsedEntityDeclHandler (void *pUserData, const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
int  _ExpatExternalEntityRefHandler        (XML_Parser parser, const XML_Char *context, const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId);
void _ExpatDefaultHandler                  (void *pUserData, const XML_Char *s, int len);
void _ExpatDefaultHandlerExpand            (void *pUserData, const XML_Char *s, int len);
void _ExpatCharacterDataHandler            (void *pUserData, const XML_Char *s, int len);
void _ExpatStartDoctypeDeclHandler         (void *pUserData, const XML_Char *doctypeName);
void _ExpatEndDoctypeDeclHandler           (void *pUserData);
void _ExpatStartCdataSectionHandler        (void *pUserData);
void _ExpatEndCdataSectionHandler          (void *pUserData);
int  _ExpatNotStandaloneHandler            (void *pUserData);
int  _ExpatUnknownEncodingHandler          (void *pUserData, const XML_Char *name, XML_Encoding *info);
/* functions called by UnknownEncodingHandler */
int  _ExpatUnknownEncodingConvert          (void *pUserData, const char *s);
void _ExpatUnknownEncodingRelease          (void *pUserData);



/* bye ;-) */
#ifdef __cplusplus
}
#endif

#endif	/* not RexxExpat_INCLUDED */

